"""
This module provides alternative methods of performing the 2d
projection needed, they will use the same interface and should be
interchangeable.

All functions will take as inputs a transform and an array of AABB
using floating point numbers. The result will be quantized into
integers.
"""
from __future__ import print_function
import ctypes
import numpy as np
import os
import sys

import distutils.sysconfig
so_ext = distutils.sysconfig.get_config_var('EXT_SUFFIX') \
        or distutils.sysconfig.get_config_var('SHLIB_SUFFIX') 
if not so_ext:
    so_ext = distutils.sysconfig.get_config_var('SO') or '.so'

def _type_lib(lib):
    from ctypes import c_void_p, c_size_t

    lib.transform_f.argtypes = [c_void_p, c_void_p, c_void_p, c_size_t, c_size_t]
    lib.transform_d.argtypes = [c_void_p, c_void_p, c_void_p, c_size_t, c_size_t]

    if hasattr(lib, "async_transform_f_start"):
        #assume all the async interface
        lib.async_transform_f_start.argtypes = [c_void_p, c_void_p, 
                                                c_size_t, c_size_t]
        lib.async_transform_f_start.restype = c_void_p
        lib.async_transform_f_end.argtypes = [c_void_p]
        lib.async_transform_f_next.argtypes = [c_void_p, c_void_p, c_void_p]
        lib.async_transform_d_start.argtypes = [c_void_p, c_void_p, 
                                                c_size_t, c_size_t]
        lib.async_transform_d_start.restype = c_void_p
        lib.async_transform_d_end.argtypes = [c_void_p]
        lib.async_transform_d_next.argtypes = [c_void_p, c_void_p, c_void_p]
_lib_filename = 'transform{0}'.format(so_ext)
_lib = ctypes.CDLL(os.path.join(os.path.dirname(__file__), _lib_filename))
_type_lib(_lib)

try:
    _lib_filename = "transform_libdispatch{0}".format(so_ext)
    _lib_dispatch = ctypes.CDLL(os.path.join(os.path.dirname(__file__), _lib_filename))
    _type_lib(_lib_dispatch)
except OSError:
    _lib_dispatch = _lib

if sys.version_info.major > 2:
    mk_buff = ctypes.pythonapi.PyMemoryView_FromMemory
else:
    mk_buff = ctypes.pythonapi.PyBuffer_FromMemory
    mk_buff.restype = ctypes.py_object

def _projectRects(viewxform, inputs, outputs, use_dispatch = False):
    if (inputs.flags.f_contiguous): 
        inputs = inputs.T
        outputs = outputs.T

    assert(len(inputs.shape) == 2 and inputs.shape[0] == 4)
    assert(inputs.shape == outputs.shape)

    use_lib = _lib_dispatch if use_dispatch else _lib

    if inputs.dtype == np.float64:
        inputtype = ctypes.c_double
        func = use_lib.transform_d
    elif inputs.dtype == np.float32:
        inputtype = ctypes.c_float
        func = use_lib.transform_f
    else:
        raise TypeError("ProjectRects only works for np.float32 and np.float64 inputs")

    assert(outputs.dtype == np.int32)

    t = ctypes.POINTER(inputtype)
    cast = ctypes.cast
    c_xforms = (inputtype * 4)(*viewxform)
    c_inputs = (t * 4)(*(cast(inputs[i].ctypes.data, t) 
                         for i in range(0, 4)))
    c_outputs = (t* 4)(*(cast(outputs[i].ctypes.data, t)
                         for i in range(0,4)))
    func(ctypes.byref(c_xforms),
         ctypes.byref(c_inputs),
         ctypes.byref(c_outputs),
         0,
         inputs.shape[1])
        

def _projectRectsGenerator(viewxform,
                           inputs,
                           chunk_size,
                           use_dispatch = False):
    if (inputs.flags.f_contiguous):
        inputs = inputs.T
        # FIXME: outputs is undefined?
        outputs = outputs.T

    assert(len(inputs.shape) == 2 and inputs.shape[0] == 4)
    result_size = inputs.shape[1]
    chunk_size = min(result_size, chunk_size)

    use_lib = _lib_dispatch if use_dispatch else _lib
    if inputs.dtype == np.float64:
        inputtype = ctypes.c_double
        func = use_lib.transform_d
        has_async = hasattr(use_lib, "async_transform_d_start")
    elif inputs.dtype == np.float32:
        inputtype = ctypes.c_float
        func = use_lib.transform_f
        has_async = hasattr(use_lib, "async_transform_f_start")
    else:
        raise TypeError("ProjectRects only works for np.float32 and np.float64 inputs")

    t = ctypes.POINTER(inputtype)
    cast = ctypes.cast
    c_xforms = (inputtype * 4)(*viewxform)
    c_inputs = (t*4)(*(cast(inputs[i].ctypes.data, t)
                       for i in range(0,4)))

    if not has_async:
        print("no async")
        # simple case: no need to pipeline
        outputs = np.empty((4, chunk_size), dtype=np.int32)
        c_outputs = (t*4)(*(cast(outputs[i].ctypes.data, t)
                            for i in range(0, 4)))
        offset = 0
        while offset != result_size:
            curr_size = min(result_size - offset, chunk_size)
            func(ctypes.byref(c_xforms),
                 ctypes.byref(c_inputs),
                 ctypes.byref(c_outputs),
                 offset,
                 curr_size)
            offset += curr_size
            yield outputs[0:curr_size]
    else:
        print("async")
        if inputs.dtype == np.float64:
            start_function = use_lib.async_transform_d_start
            end_function  = use_lib.async_transform_d_end
            next_function  = use_lib.async_transform_d_next
        elif inputs.dtype == np.float32:
            start_function = use_lib.async_transform_f_start
            end_function  = use_lib.async_transform_f_end
            next_function  = use_lib.async_transform_f_next
        else:
            raise TypeError("ProjectRects only works for np.float32 and np.float64 inputs")

        buff = ctypes.c_void_p(0)
        count = ctypes.c_size_t(0)
        total_size = chunk_size * 4 * 4
        token = start_function(ctypes.byref(c_xforms), ctypes.byref(c_inputs),
                               ctypes.c_size_t(result_size),
                               ctypes.c_size_t(chunk_size))

        while True:
            next_function(token, ctypes.byref(buff), ctypes.byref(count))
            if count.value == 0:
                break

            yield np.frombuffer(mk_buff(buff, total_size, PyBuff_READ),
                                dtype='i4').reshape((4,chunk_size))[:,0:count.value]

        end_function(token)

# testing code starts here

def _project_element(viewxform, inputs, output):
    tx, ty, sx, sy = viewxform
    x, y, w, h = inputs
    x2 = x + w
    y2 = y + h
    np.floor(sx * x + tx, out=output[0,:])
    np.floor(sy * y + ty, out=output[1,:])
    np.floor(sx * x2 + tx, out=output[2,:])
    np.floor(sy * y2 + ty, out=output[3,:])


def report_diffs(a, b, name):
    last_dim = a.shape[1]
    if not np.allclose(a, b):
        for i in xrange(1, last_dim):
            if not np.allclose(a[:,i], b[:,i]):
                print('%s::%d fails \nc:\n%s != %s\n' % 
                      (name, i, str(a[:,i]), str(b[:,i])))


def simple_test():
    from time import time

    use_shape = (4, 10**8)
    chunksize = 10**4
    mock_in = np.random.random(use_shape)
    xform = [3.0, 4.0, 2.0, 2.0]

    t = time()
    check_sum = 0
    for arr in _projectRectsGenerator(xform, mock_in, chunksize):
        check_sum += arr.shape[1]
        np.random.random(1000)
    t = time() - t
    print("checksum (_projectRectsGenerator) took %f ms" % (t*1000))
    chk1 = check_sum

    t = time()
    check_sum = 0
    for arr in _projectRectsGenerator(xform, mock_in, chunksize, use_dispatch = True):
        check_sum += arr.shape[1]
        np.random.random(1000)

    t = time() - t
    print("checksum (_projectRectsGenerator libdispatch) took %f ms" % (t*1000))
    chk2 = check_sum

    t = time()
    out = np.empty(use_shape, dtype=np.int32)
    res = _projectRects(xform, mock_in, out)
    t = time() - t
    print("checksum (_projectRects) took %f ms" % (t*1000))

    if not (chk1 == chk2 == out.shape[1]):
        print('checksums diverge')
        print('%s == %s' % ('chk1', chk1)) 
        print('%s == %s' % ('chk2', chk2))

    t = time()
    _projectRects(xform, mock_in, out, use_dispatch=True)
    t = time() - t
    out0 = np.copy(out)
    print("c version - libdispatch (double) took %f ms" % (t*1000))

    t = time()
    _projectRects(xform, mock_in, out)
    t = time() - t
    out1 = np.copy(out)
    print("c version (double) took %f ms" % (t*1000))

    t = time()
    _project_element(xform, mock_in, out)
    t = time() - t
    out2 = np.copy(out)
    print("numpy version (double) took %f ms" % (t*1000))

    mock_in = mock_in.astype(np.float32)

    t = time()
    _projectRects(xform, mock_in, out, use_dispatch=True)
    t = time() - t
    out3 = np.copy(out)
    print("c version - libdispatch (single) took %f ms" % (t*1000))

    t = time()
    _projectRects(xform, mock_in, out)
    t = time() - t
    out4 = np.copy(out)
    print("c version (single) took %f ms" % (t*1000))

    t = time()
    _project_element(xform, mock_in, out)
    t = time() - t
    out5 = np.copy(out)
    print("numpy version (single) took %f ms" % (t*1000))

    
    report_diffs(out0, out2, "libdispatch (double)")
    report_diffs(out1, out2, "plain C (double)")
    report_diffs(out3, out5, "libdispatch (single)")
    report_diffs(out4, out5, "plain C (single)")


if __name__ == '__main__':
    simple_test()



from __future__ import print_function, division, absolute_import


def bressenham(canvas, line, val):
    """
    based on 'optimized and simplified' version at
    http://en.wikipedia.org/wiki/Bresenham's_line_algorithm
    """
    (x0, y0, x1, y1) = line

    dx = abs(x1-x0)
    dy = abs(y1-y0)
    err = dx - dy

    sx = 1 if x0 < x1 else -1
    sy = 1 if y0 < y1 else -1

    while True:
        canvas[y0, x0] = val

        if x0 == x1 and y0 == y1:
            break

        e2 = err * 2
        if e2 > -dy:
            err = err - dy
            x0 = x0 + sx

        if e2 < dx:
            err = err + dx
            y0 = y0 + sy



""" Core abstract rendering abstractions. This includes the main drivers of
execution and the base clases for shared data representations.
"""
from __future__ import print_function, division, absolute_import
from six.moves import range

import numpy as np
import abstract_rendering.geometry as geometry
import abstract_rendering.glyphset as glyphset


# ------------------- Basic process function --------------------------------
def render(glyphs, info, aggregator, shader, screen, vt):
    """
    Render a set of glyphs to the described canvas.

    * glyphs -- Glyphs to render
    * info -- For each glyph, the piece of information that will be aggregated
    * aggregator -- Combines a set of info values into a single aggregate value
    * shader -- Converts aggregates to other aggregates (often colors)
    * screen -- (width,height) of the canvas
    * vt -- View transform (converts canvas to pixels)
    """
    projected = glyphs.project(vt)
    aggregates = aggregator.aggregate(projected, info, screen)
    # TODO: Add shader specialization here
    rslt = shader(aggregates)
    return rslt


# -------------------------  Aggregators and related utilities ----------------
class Aggregator(object):
    out_type = None
    in_type = None
    identity = None

    def aggregate(self, glyphset, info, screen):
        """
        Produce a set of aggregates

        glyphset -- glyphs to process
        screen -- (width, height) of the output grid
        info -- info function to invoke
        """

        raise NotImplementedError()

    def rollup(self, *vals):
        """
        Combine multiple sets of aggregates.

        * vals - list of numpy arrays with type out_type
        """
        raise NotImplementedError()


class GlyphAggregator(Aggregator):
    """
    Aggregator tha tworks on one glyph at a time.

    Aggregators need to eventually process all glyphs.
    This class provides on workflow for realzing that.
    Each glyph is turned into its own set of aggregates,
    then combine dinto a larger set of aggregates for the
    whole glyphset.

    High-level overview of the control flow:
        * 'allocate' is used to make an empty set of aggregates
          for the whole glyphset
        * 'aggregate' calls 'combine' to include a single glyph
          into that allocated set of aggregates.
        * 'aggregate' repeats until all glyphs have been processed
        * 'glyphAggregates' is a utility for combine
          to convert a glyph into a set of aggregates.  Most instances
          of 'combine' call 'glyphAggregates' though it is not always
          required

    Sub-classes need to implement allocate and combine.
    """

    def allocate(self, glyphset, screen):
        """
        Create an array suitable for processing the passed dataset
        into the requested grid size.

        * glyphset - The points that will be processed (already projected)
        * screen -- The size of the bin-grid to produce
        """
        raise NotImplementedError()

    def combine(self, existing, points, shapecode, val):
        """Add a new point to an existing set of aggregates.

        * existing - out_type numpy array, aggregate values for all glyphs seen
        * points - points that define a shape
        * shapecode - Code that determines how points are interpreted
        * val -- Info value associated with the current set of points
        """
        raise NotImplementedError()

    def aggregate(self, glyphset, info, screen):
        # TODO: vectorize pretty much this whole method...
        (width, height) = screen

        # co-iterating on number of points in case glyphset.data() is a non-length-carrying placeholder
        # TODO: Should the default placeholder carry length?
        infos = [info(data) for (data, _)
                 in zip(glyphset.data(), range(len(glyphset.points())))]
        aggregates = self.allocate(glyphset, screen)
        for idx, points in enumerate(glyphset.points()):
            self.combine(aggregates,
                         points,
                         glyphset.shaper.code,
                         infos[idx])
        return aggregates

    def glyphAggregates(self, glyph, shapeCode, val, default):
        """Create a set of aggregates for a single glyph. The set of aggregates will be
           tight to the bound box of the shape but may not be completely filled
           (thus the need for both 'val' and 'default').

           * glyph -- Points that define the glyph
           * shapeCode -- Code that indicates how to interpret the glyph
           * val -- Value to place in bins that are hit by the shape
           * default -- Value to place in bins not hit by the shape
        """

        def scalar(array, val):
            array.fill(val)

        def nparray(array, val):
            array[:] = val

        if type(val) == np.ndarray:
            fill = nparray
            extShape = val.shape
        else:
            fill = scalar
            extShape = ()

        # TODO: These are selectors...rename and move this somewhere else
        if shapeCode == glyphset.ShapeCodes.POINT:
            array = np.copy(val)  # TODO: Not sure this is always an array...verify
        elif shapeCode == glyphset.ShapeCodes.RECT:
            array = np.empty((glyph[3]-glyph[1], glyph[2]-glyph[0])+extShape,
                             dtype=np.int32)
            fill(array, val)
        elif shapeCode == glyphset.ShapeCodes.LINE:
            array = np.empty((glyph[3]-glyph[1], glyph[2]-glyph[0])+extShape,
                             dtype=np.int32)
            fill(array, default)
            glyph = [0, 0, array.shape[1]-1, array.shape[0]-1]  # Translate shape to be in the corner of the update canvas
            geometry.bressenham(array, glyph, val)

        return array


# ---------------------- Shaders and related utilities --------------------
class Shader(object):
    """Shaders take grids and analize them.
       This interface asserts that instances are callable
       and accept a grid as their input.
    """

    def __add__(self, other):
        """Extend this shader by executing another transfer in sequence."""
        if (not isinstance(other, Shader)):
                raise TypeError("Can only extend with Shaders.  Received a {0}"
                                .format(str(type(other))))
        return Seq(self, other)


class ShapeShader(Shader):
    "Convert a grid into a set of shapes (instead of another grid)."

    def fuse(self, grid):
        "Convert aggregates grid into geometry"
        raise NotImplementedError

    def __call__(self, grid):
        return self.fuse(grid)


# TODO: Add specialization to Shaders....
class CellShader(Shader):
    """Cell shaders take a grid and produce a new grid."""

    def shade(self, grid):
        """Execute the actual data shader operation."""
        raise NotImplementedError

    def __call__(self, grid):
        """Execute shading (by default)."""
        return self.shade(grid)


class Seq(Shader):
    "Shader that does a sequence of shaders."

    def __init__(self, *args):
        self._parts = args

    def __add__(self, other):
        if (other is None):
            return self
        elif not isinstance(self._parts[-1], CellShader):
            raise ValueError("Cannot extend: Sequence terminated by non-CellShader.")
        elif (not isinstance(other, Shader)):
            raise TypeError("Can only extend with Shaders. Received a "
                            .format(str(type(other))))
        return Seq(*(self._parts + (other,)))

    def __call__(self, grid):
        for t in self._parts:
            grid = t(grid)
        return grid


class SequentialShader(Shader):
    "Data shader that does non-vectorized per-pixel shading."

    def _pre(self, grid):
        "Executed exactly once before pixelfunc is called on any cell. "
        pass

    def __call__(self, grid):
        """Execute shading."""
        return self.shade(grid)

    def cellfunc(grid, x, y):
        """
        This method will be called for each pixel in the outgrid.
        Must be implemented in subclasses.
        """
        raise NotImplementedError

    def makegrid(self, grid):
        """Create an output grid.

           Default implementation creates one of the same width/height
           of the input suitable for colors (dept 4, unit8).
        """
        (width, height) = grid.shape[0], grid.shape[1]
        return np.ndarray((width, height, 4), dtype=np.uint8)

    def shade(self, grid):
        """Access each element in the out grid sequentially"""
        outgrid = self.makegrid(grid)
        self._pre(grid)
        (height, width) = outgrid.shape
        for x in range(0, width):
            for y in range(0, height):
                outgrid[y, x] = self.cellfunc(grid, x, y)

        return outgrid



from __future__ import print_function, division, absolute_import


class EmptyList(object):
    """
    Utility that can be numerically indexed, but
    always returns None.

    If a no length or a negative length are passed at construction,
    the list will ALWAYS return None.

    If a non-negative length is passsed at construction,
    an indexed 0 <= index < length will return None.
    Others raise an IndexError
    """

    def __init__(self, length=-1):
        self.length = length

    def __getitem__(self, idx):
        if self.length < 0:
            return None

        if idx >= self.length or idx < 0:
            raise IndexError

        return None


class Color(list):
    def __init__(self, r, g, b, a):
        list.__init__(self, [r, g, b, a])
        self.r = r
        self.g = g
        self.b = b
        self.a = a

        if ((r > 255 or r < 0)
                or (g > 255 or g < 0)
                or (b > 255 or b < 0)
                or (a > 255 or a < 0)):
            raise ValueError


def zoom_fit(screen, bounds, balanced=True):
    """What affine transform will zoom-fit the given items?
         screen: (w,h) of the viewing region
         bounds: (x,y,w,h) of the items to fit
         balance: Should the x and y scales match?
         returns: [translate x, translate y, scale x, scale y]
    """
    (sw, sh) = screen
    (gx, gy, gw, gh) = bounds
    x_scale = sw/gw
    y_scale = sh/gh
    if (balanced):
        x_scale = min(x_scale, y_scale)
        y_scale = x_scale
    return [-gx*x_scale, -gy*y_scale, x_scale, y_scale]



"""
Utilities that apply across a broad variety of categories.
"""

from __future__ import print_function, division, absolute_import
import abstract_rendering.core as core


class Id(core.CellShader):
    """ Return the input unchanged.

    This DOES NOT make a copy of the input
    It is usually used a zero-cost placeholder.
    """

    def makegrid(self, grid):
        return grid

    def shade(self, grid):
        return grid



from __future__ import print_function, division, absolute_import
from six.moves import map
import numpy as np
import re
from abstract_rendering.fast_project import _projectRects
import six

def enum(**enums): return type('Enum', (), enums)
ShapeCodes = enum(POINT=0, LINE=1, RECT=2)
# TODO: Add shapecodes: CIRCLE, POINT_RECT, POINT_CIRCLE, etc


class Glyphset(object):
    """
    shaper + shape-params + associated data ==> Glyphset
    This reference implementation uses numpy arrays for the backing store.

    fields:
       _points : Points held by this glyphset
       _data : Data associated with the points (basis for 'info' values).
       shapecode: Shapecode that tells how to interpret _points
       Note: _points[x] should associate with data[x]
    """
    _points = None
    _data = None
    shaper = None

    def __init__(self, points, data, shaper, colMajor=False):
        self._points = points
        self._data = data
        self.shaper = shaper
        self.shaper.colMajor = colMajor  # HACK: This modification is bad form...would rather do a copy

    # TODO: Add the ability to get points m...n
    def points(self):
        """Returns the set of [x,y,w,h] points for this glyphset.
           Access to raw data is through _points.
        """

        if (type(self.shaper) is Literals):
            if type(self._points) is list:
                return np.array(self._points, order="F")
            elif type(self._points) is np.ndarray:
                return self._points
            else:
                ValueError("Unhandled (literal) points type: %s"
                           % type(self._points))
        else:
            # TODO: Setup the shaper utilities to go directly to fortran order
            return np.array(self.shaper(self._points), order="F")

    def project(self, viewxform):
        """Project the points found in the glyphset according to the view transform.

        viewxform -- convert canvas space to pixel space [tx,ty,sx,sy]
        returns a new glyphset with projected points and associated info values
        """
        points = self.points()
        out = np.empty_like(points, dtype=np.int32)
        _projectRects(viewxform, points, out)

        # Ensure visilibity, make sure w/h are always at least one
        # TODO: There is probably a more numpy-ish way to do this...
        for i in xrange(0, out.shape[0]):
            if out[i, 0] == out[i, 2]:
                out[i, 2] += 1
            if out[i, 1] == out[i, 3]:
                out[i, 3] += 1

        return Glyphset(out, self.data(), Literals(self.shaper.code))

    def data(self):
        return self._data

    def bounds(self):
        """Compute bounds of the glyph-set.  Returns (X,Y,W,H)

           Assumes a 'simple' layout where the min and max input values
           determine the min and max positional values.
        """
        minX = float("inf")
        maxX = float("-inf")
        minY = float("inf")
        maxY = float("-inf")
        for g in self.points():
            (x, y, w, h) = g

            minX = min(minX, x)
            maxX = max(maxX, x+w)
            minY = min(minY, y)
            maxY = max(maxY, y+h)
        return (minX, minY, maxX-minX, maxY-minY)
#        points = self.points();
#        minX=points[:,0].min()
#        maxX=points[:,0].max()
#        minY=points[:,1].min()
#        maxY=points[:,1].max()
#        maxW=points[:,2].max()
#        maxH=points[:,3].max()
#
#        shapes = self.shaper([[minX,maxX],[minY,maxY],[maxW,maxW],[maxH,maxH]])
#        minX = shapes[0][0]
#        minY = shapes[0][1]
#        maxX = shapes[1][0]
#        maxY = shapes[1][1]
#        width = shapes[1][2]
#        height = shapes[1][3]
#
#        return (minX, minY, maxX+width, maxY+height)


# Shapers.....
class Shaper(object):
    fns = None  # List of functions to apply
    code = None
    colMajor = False

    # TODO: When getting subsets of the data out of glyphset.points(), remove this colMajor and handle it up in glyphset instead
    def __call__(self, vals):
        if not self.colMajor:
            shapes = [list(map(lambda f: f(val), self.fns)) for val in vals]
        else:
            shapes = [list(map(lambda f: f(val), self.fns)) for val in zip(*vals)]

        return shapes


class Literals(Shaper):
    """Optimization marker, tested in Glyphset to skip conversions.
       Using this class asserts that the _points value of the Glyphset
       is the set of points to feed into a geometry renderer, and thus
       no projection from data space to canvas space is required.
       Use with caution...
    """
    def __init__(self, code):
        self.code = code

    def __call__(self, vals):
        return vals


class ToRect(Shaper):
    """Creates rectangles using functions to build x,y,w,h."""
    code = ShapeCodes.RECT

    def __init__(self, tox, toy, tow, toh):
        self.fns = [tox, toy, tow, toh]


class ToLine(Shaper):
    """Creates lines using functions to build x1,y1,x2,y2"""
    code = ShapeCodes.LINE

    def __init__(self, tox1, toy1, tox2, toy2):
        self.fns = [tox1, toy1, tox2, toy2]


class ToPoint(Shaper):
    """Create a single point, using functions to build x,y,0,0
       Can accept more than just tox/toy arguments, but ignores them.
    """
    code = ShapeCodes.POINT

    def __init__(self, tox, toy, *args):
        self.fns = [tox, toy, lambda x: 0, lambda x: 0]


# Utilities for shapers....
def const(v):
    """Create a function that always returns a specific value.

    * v -- The value the returned function always returns
    """
    def f(a):
        return v
    return f


def item(i):
    """
    Get items out of a collection.
    Suiteable for use with numerically indexed (i.e., array)
    or object-indexed (i.e., dictionary) sources.

    * i -- The item parameter that will be used
    """
    def f(a):
        return a[i]
    return f


def idx(i):
    "The same as item, but a common name when using numeric indexes"
    return item(i)


def load_csv(filename, skip, xc, yc, vc, width, height, shapecode):
    """Turn a csv file into a glyphset.

    This is a fairly naive regulary-expression based parser
    (it doesn't handle quotes, blank lines or much else).
    It is useful for getting simple datasets into the system.
    """
    source = open(filename, 'r')
    glyphs = []
    data = []

    if shapecode is ShapeCodes.POINT:
        width = 0
        height = 0

    for i in range(0, skip):
        source.readline()

    for line in source:
        line = re.split("\s*,\s*", line)
        x = float(line[xc].strip())
        y = float(line[yc].strip())
        v = float(line[vc].strip()) if vc >= 0 else 1
        g = [x, y, width, height]
        glyphs.append(g)
        data.append(v)

    source.close()
    return Glyphset(glyphs, data, Literals(shapecode))


def load_hdf(filename, node, xc, yc, vc, width, height, shapecode):
    """
    Load a node from an HDF file.

    filename : HDF file to load
    node: Path to relevant HDF table
    xc: Name/index of the x column
    yc: Name/index of the y column
    vc: Name/index of the value column (if applicable)
    cats: List of expected categories.
        If cats is an empty list, a coding will be automatically generated
        Any value not on the list will be assigned category equal to
        list length. Cats is ignored if vc is not supplied.
    """
    import pandas as pd
    table = pd.read_hdf(filename, node)
    x = np.array(table[xc])
    y = np.array(table[yc])
    w = np.array([width] * len(x))
    h = np.array([height] * len(x))
    points = np.vstack([x, y, w, h]).T
    data = table[vc] if vc else ([1] * len(x))
    print("Loaded %d items" % len(x))

    return Glyphset(points, data, Literals(shapecode))



from __future__ import print_function, division, absolute_import
from six.moves import reduce

import numpy as np
import math
import abstract_rendering.core as core
import abstract_rendering.util as util


# ----------- Aggregators -----------
class Count(core.GlyphAggregator):
    """Count the number of items that fall into a particular grid element."""
    out_type = np.int32
    identity = 0

    def allocate(self, glyphset, screen):
        (width, height) = screen
        return np.zeros((height, width), dtype=self.out_type)

    def combine(self, existing, glyph, shapecode, val):
        update = self.glyphAggregates(glyph, shapecode, 1, self.identity)
        existing[glyph[1]:glyph[3], glyph[0]:glyph[2]] += update

    def rollup(self, *vals):
        return reduce(lambda x, y: x+y,  vals)


class Sum(core.GlyphAggregator):
    """Count the number of items that fall into a particular grid element."""
    out_type = np.int32
    identity = 0

    def allocate(self, glyphset, screen):
        (width, height) = screen
        return np.zeros((height, width), dtype=self.out_type)

    def combine(self, existing, glyph, shapecode, val):
        update = self.glyphAggregates(glyph, shapecode, val, self.identity)
        existing[glyph[1]:glyph[3], glyph[0]:glyph[2]] += update

    def rollup(self, *vals):
        return reduce(lambda x, y: x+y,  vals)


# -------------- Shaders -----------------
class Floor(core.CellShader):
    def shade(self, grid):
        return np.floor(grid)


class Interpolate(core.CellShader):
    """Interpolate between two numbers.
         Projects the input values between the low and high values passed.
         The Default is 0 to 1.
         Empty values are preserved (default is np.nan).
    """
    def __init__(self, low=0, high=1, empty=np.nan):
        self.low = low
        self.high = high
        self.empty = empty

    def shade(self, grid):
        # TODO: Gracefully handle if the whole grid is empty
        mask = (grid == self.empty)
        min = grid[~mask].min()
        max = grid[~mask].max()
        span = float(max-min)
        percents = (grid-min)/span
        return self.low + (percents * (self.high-self.low))


class Power(core.CellShader):
    """Raise to a power.    Power may be fracional."""
    def __init__(self, pow):
        self.pow = pow

    def shade(self, grid):
        return np.power(grid, self.pow)


class Cuberoot(Power):
    def __init__(self):
        super(Cuberoot, self).__init__(1/3.0)


class Sqrt(core.CellShader):
    def shade(self, grid):
        return np.sqrt(grid)


class Spread(core.SequentialShader):
    """
    Spreads the values out in a regular pattern.

    * factor : How far in each direction to spread

    TODO: Currently only does square spread.  Extend to other shapes.
    TODO: Restricted to numbers right now...implement corresponding thing
    for categories...might be 'generic'
    """

    def __init__(self, up=1, down=1, left=1, right=1, factor=np.NaN):
        if np.isnan(factor):
            self.up = up
            self.down = down
            self.left = left
            self.right = right
        else:
            self.up = factor
            self.down = factor
            self.left = factor
            self.right = factor

    def makegrid(self, grid):
        height = grid.shape[0]
        width = grid.shape[1]
        others = grid.shape[2:]

        height = height + self.up + self.down
        width = width + self.left + self.right

        return np.zeros((height, width) + others, dtype=grid.dtype)

    def cellfunc(self, grid, x, y):
        (height, width) = grid.shape
        minx = max(0, x-self.left-self.right)
        maxx = min(x+1, width)
        miny = max(0, y-self.up-self.down)
        maxy = min(y+1, height)

        parts = grid[miny:maxy, minx:maxx]
        return parts.sum()


class BinarySegment(core.CellShader):
    """
    Paint all pixels with aggregate value above divider one color
    and below the divider another.  Divider is part of the 'high' region.

    TODO: Extend so out can be something other than colors
    """
    in_type = (1, np.number)
    out_type = (4, np.int32)

    def __init__(self, low, high, divider):
        self.high = high
        self.low = low
        self.divider = float(divider)

    def shade(self, grid):
        (width, height) = grid.shape[0], grid.shape[1]
        outgrid = np.ndarray((width, height, 4), dtype=np.uint8)
        mask = (grid >= self.divider)
        outgrid[mask] = self.high
        outgrid[~mask] = self.low
        return outgrid


class InterpolateColors(core.CellShader):
    """
    High-definition interpolation between two colors.
    Zero-values are treated separately from other values.

    TODO: Remove log, just provide a shader to pre-transform the values
    TODO: Can this be combined with 'Interpolate'? Detect type at construction

    * low -- Color ot use for lowest value
    * high -- Color to use for highest values
    * log -- Set to desired log base to use log-based interpolation
                     (use True or "e" for base-e; default is False)
    * reserve -- color to use for empty cells
    """
    in_type = (1, np.number)
    out_type = (4, np.int32)

    def __init__(self,
                 low, high,
                 log=False,
                 reserve=util.Color(255, 255, 255, 255),
                 empty=np.nan):
        self.low = low
        self.high = high
        self.reserve = reserve
        self.log = log
        self.empty = empty

    # TODO: there are issues with zeros here....
    def _log(self,  grid):
        mask = (grid == self.empty)
        min = grid[~mask].min()
        max = grid[~mask].max()

        grid[mask] = 1
        if (self.log == 10):
            min = math.log10(min)
            max = math.log10(max)
            span = float(max-min)
            percents = (np.log10(grid)-min)/span
        elif (self.log == math.e or self.log):
            min = math.log(min)
            max = math.log(max)
            span = float(max-min)
            percents = (np.log(grid)-min)/span
        elif (self.log == 2):
            min = math.log(min, self.log)
            max = math.log(max, self.log)
            span = float(max-min)
            percents = (np.log2(grid)-min)/span
        else:
            rebase = math.log(self.log)
            min = math.log(min, self.log)
            max = math.log(max, self.log)
            span = float(max-min)
            percents = ((np.log(grid)/rebase)-min)/span

        grid[mask] = 0

        colorspan = (np.array(self.high, dtype=np.uint8)
                     - np.array(self.low, dtype=np.uint8))

        outgrid = (percents[:, :, np.newaxis]
                   * colorspan[np.newaxis, np.newaxis, :]
                   + np.array(self.low, dtype=np.uint8)).astype(np.uint8)
        outgrid[mask] = self.reserve
        return outgrid

    def _linear(self, grid):
        mask = (grid == self.empty)
        min = grid[~mask].min()
        max = grid[~mask].max()
        span = float(max-min)
        percents = (grid-min)/span

        colorspan = (np.array(self.high, dtype=np.int32)
                     - np.array(self.low, dtype=np.int32))
        outgrid = (percents[:, :, np.newaxis]
                   * colorspan[np.newaxis, np.newaxis, :]
                   + np.array(self.low, dtype=np.uint8)).astype(np.uint8)
        outgrid[mask] = self.reserve
        return outgrid

    def shade(self, grid):
        if (self.log):
            return self._log(grid)
        else:
            return self._linear(grid)



from __future__ import print_function, division, absolute_import
import numpy as np
import abstract_rendering.core as core
from abstract_rendering._cntr import Cntr


class Contour(core.ShapeShader):
    def __init__(self, x_range=None, y_range=None, levels=5, points=True):
        """
         x/y_range arguments determine the data values that correspond
         to the max/min values on the axes of the input grid.  It is assumed
         that the grid is linear between the two.

         levels as a scalar determines how many levels will be built
         levels as a list determines where the levels are built

         points -- Indicates if it should return values
                    as [[(x,y)...], [(x,y)...]] (default, True)
                    or as [[x,x...], [x,x...],[y,y...][y,y...]] (False)

        """

        self.x_range = x_range
        self.y_range = y_range
        self.levels = levels
        self.points = points

    def fuse(self, grid):
        x_range = self.x_range if self.x_range else (0, grid.shape[1]-1)
        y_range = self.y_range if self.y_range else (0, grid.shape[0]-1)

        xs = np.linspace(x_range[0], x_range[1], num=grid.shape[1])
        ys = np.linspace(y_range[0], y_range[1], num=grid.shape[0])
        xg, yg = np.meshgrid(xs, ys)

        c = Cntr(xg, yg, grid)

        if isinstance(self.levels, list):
            levels = self.levels
        else:
            levels = Contour.nlevels(grid, self.levels)

        isos = dict()
        for level in levels:
            points = c.trace(level, points=self.points)
            points = [[], []] if len(points) == 0 else points
            isos[level] = points

        return isos

    @classmethod
    def nlevels(cls, grid, n):
        "Given grid of values, pick out n values for iso contour levels"

        # The contouring library outlines places below the given value.
        # We do n+2 and drop the first and last values because they are
        # empty contours.
        return np.linspace(grid.min(), grid.max(), n+2)[1:-1]



"""
Tools for working with counts from multiple categories of data at once.

Categories are modeled as stakced 2D arrays.    Each category is in its
own slice of the stack.
"""
from __future__ import print_function, division, absolute_import
from six.moves import reduce
import numpy as np
from math import log
import abstract_rendering.util as util
import abstract_rendering.core as core


# ------------ Aggregators -------------------
class CountCategories(core.GlyphAggregator):
    """
    Count the number of items that fall into a particular grid element.

    To operateo correctly, allocate must be called with infos representing
    all possible categories; that set of categories is used in combine
    to determine where to place values.  It is, therefore, important that
    any call to combine must in some way correspond to the most recent
    call to allocate.

    TODO: This seems convoluted...is there a better way?
    """
    out_type = np.int32
    identity = np.asarray([0])
    cats = None

    def allocate(self, glyphset, screen):
        """Allocates one array slice for each unique info passed.
           Output array shape is (#cats, height, width).

        """
        (width, height) = screen
        self.cats = np.unique(glyphset.data())
        return np.zeros((height, width, len(self.cats)), dtype=self.out_type)

    def combine(self, existing, points, shapecode, val):
        entry = np.zeros(len(self.cats))
        idx = np.nonzero(self.cats == val)[0][0]
        entry[idx] = 1
        update = self.glyphAggregates(points, shapecode, entry, self.identity)
        existing[points[1]:points[3], points[0]:points[2], :] \
                += update

    def rollup(self, *vals):
        """NOTE: Assumes co-registration of categories..."""
        return reduce(lambda x, y: x+y,  vals)


# ----------- Shaders -----------------------
class ToCounts(core.CellShader):
    """Convert from count-by-categories to just raw counts.
       Then data shader functions from the count module can be used.
    """
    @staticmethod
    def shade(grid, dtype=None):
        dtype = (grid.dtype if dtype is None else dtype)
        return grid.sum(axis=2, dtype=dtype)


class NonZeros(core.CellShader):
    "How many non-zero categories are there?"
    def shade(self, grid):
        mask = grid != 0
        nonzero = np.zeros_like(grid)
        nonzero[mask] = 1
        return nonzero.sum(axis=2)


class Ratio(core.CellShader):
    """
    Compute the ratio of one category to the sum of all cateogires.
    Any cell with sum of zero is propogated as a 'zero' in the sum
     * focus: The "one cateogry"
              Will pick the local 'max' cateogry if set to a negative
    """
    def __init__(self, focus=-1):
        self.focus = focus

    def shade(self, grid):
        if self.focus < 0:
            plane = grid.max(axis=2)
        else:
            plane = grid[:, :, self.focus]

        sum = grid.sum(axis=2)
        mask = (sum != 0)
        ratio = plane/sum
        ratio[~mask] = 0
        return ratio


class Select(core.CellShader):
    """Get the counts from just one category.

       Operates by taking a single plane of the count of categories.

       TODO: Consider changing shade to take a wrapper 'grid' that can carry
             info like a category-label-to-grid-slice mapping....
    """

    def __init__(self, slice):
        """slice -- Which slice of the aggregates grid should be returned"""
        self.slice = slice

    def shade(self, aggregates):
        return aggregates[:, :, self.slice]


class MinPercent(core.CellShader):
    """If the item in the specified bin represents more than a certain percent
     of the total number of items, color it as "above" otherwise, "below"

     TODO: Change from idx to category label, lookup idx for 'cat' parameter
     TODO: Extend to work with more than just colors

     * cutoff -- percent value to split above and below coloring
     * cat -- integer indicating which category number to use
     * above    -- color to paint above (default is a red)
     * below    -- color to paint below (default is a blue)
     * background -- color to paint when there are no values (default is clear)
    """

    def __init__(self,
                 cutoff,
                 cat=0,
                 above=util.Color(228, 26, 28, 255),
                 below=util.Color(55, 126, 184, 255),
                 background=util.Color(255, 255, 255, 0)):

        self.cutoff = cutoff
        self.cat = cat
        self.above = np.array(above, dtype=np.uint8)
        self.below = np.array(below, dtype=np.uint8)
        self.background = np.array(background, dtype=np.uint8)

    def shade(self, grid):
        (height, width, depth) = grid.shape
        outgrid = np.empty((height, width, 4), dtype=np.uint8)

        sums = ToCounts.shade(grid, dtype=np.float32)
        maskbg = (sums == 0)
        mask = (grid[:, :, self.cat]/sums) >= self.cutoff

        outgrid[mask] = self.above
        outgrid[~mask] = self.below
        outgrid[maskbg] = self.background
        return outgrid


class HDAlpha(core.CellShader):
    def __init__(self, colors, background=util.Color(255, 255, 255, 255),
                 alphamin=0, log=False, logbase=10):
        """
        colors -- a list of colors in cateogry-order.
        alphamin -- minimum alpha value when (default is 0)
        log -- Log-based interpolate? (default is false)
        logbase -- Base to use if log is true (default is 10)
        background -- Color for empty category list (default is white)

        If C categories and N colors are presented:
            * N == C: No problems
            * N >  C: Only the first C colors are used
            * N <  C: The last color is replicated to make N == C

        TODO: Change 'colors' to a dictionary of category-to-color mapping
        TODO: mask out zero-sum regions in alpha and opaque blend
        """
        self.catcolors = np.array([np.array(v, dtype=np.uint8) for v in colors])
        self.background = np.array(background, dtype=np.uint8)
        self.alphamin = alphamin
        self.log = log
        self.logbase = logbase

    def shade(self, grid):
        sums = ToCounts.shade(grid, dtype=np.float32)
        mask = (sums != 0)

        # Ensure category count and color count match
        cats = grid.shape[2]
        catcolors = self.catcolors
        if len(catcolors) > cats:
            catcolors = catcolors[:cats]
        elif len(catcolors) < cats:
            other = catcolors[-1]
            replicated = [other] * (cats-len(catcolors))
            catcolors = np.vstack([catcolors, replicated])

        colors = HDAlpha.opaqueblend(catcolors, grid, sums)
        colors[~mask] = self.background
        HDAlpha.alpha(colors, sums, mask, self.alphamin, self.log, self.logbase)
        return colors

    # ------------------- Utilities -----------------
    @staticmethod
    def alpha(colors, sums, mask, alphamin, dolog=False, base=10):
        maxval = sums.max()

        if (dolog):
            base = log(base)
            maxval = log(maxval)/base
            sums[mask] = np.log10(sums[mask])/base

        np.putmask(colors[:, :, 3],
                   mask,
                   ((alphamin + ((1-alphamin) * (sums/maxval)))*255).astype(np.uint8))

    @staticmethod
    def opaqueblend(catcolors, counts, sums):
        weighted = (counts/sums[:, :, np.newaxis]).astype(float)
        weighted = catcolors[np.newaxis, np.newaxis, :] * weighted[:, :, :, np.newaxis]
        colors = weighted.sum(axis=2).astype(np.uint8)
        return colors



from __future__ import print_function, division, absolute_import
from six.moves import range
import numpy as np
from scipy.ndimage.filters import convolve
from abstract_rendering.fast_project import _projectRects
import abstract_rendering.glyphset as glyphset
import abstract_rendering.core as ar


class Glyphset(glyphset.Glyphset):
    # TODO: Default data is list of None (?)
    """
    points: Base array of points (x, y, w, h)
    data: Base array of data associated with points.
          points[n] is associated with data[n]
    vt: view transform
    clean_nan: Remove entries with nans in points?  Default is false.
    """
    def __init__(self, points, data, vt=(0, 0, 1, 1), clean_nan=False):
        if clean_nan:
            mask = ~np.isnan(points).any(axis=1)
            points = points[mask]
            data = data[mask]

        self._points = points
        self._data = data
        self.vt = vt
        self.shaper = glyphset.ToPoint(glyphset.idx(0), glyphset.idx(1))

        if not is_identity_transform(vt):
            self.projected = np.empty_like(points, dtype=np.int32)
            _projectRects(vt, points, self.projected)
            # points must be at least 1 wide for general compatability
            self.projected[:, 2] = self.projected[:, 0]+1
            self.projected[:, 3] = self.projected[:, 1]+1
        else:
            self.projected = points

    def data(self):
        return self._data

    def points(self):
        return self.projected

    def project(self, vt):
        """
        Project the points found in the glyphset with to the transform.

        vt -- convert canvas space to pixel space [tx,ty,sx,sy]
        returns a new glyphset with projected points and associated info values
        """
        nvt = (self.vt[0]+vt[0],
               self.vt[1]+vt[1],
               self.vt[2]*vt[2],
               self.vt[3]*vt[3])
        return Glyphset(self._points, self._data, nvt)

    def bounds(self):
        (xmax, ymax, _, _) = self.projected.max(axis=0)
        (xmin, ymin, _, _) = self.projected.min(axis=0)
        bounds = (xmin, ymin, xmax-xmin, ymax-ymin)
        return bounds


class PointCount(ar.Aggregator):
    def aggregate(self, glyphset, info, screen):
        sparse = glyphset.points()

        dense = np.histogram2d(sparse[:, 1], sparse[:, 0],
                               bins=(screen[1], screen[0]),
                               range=((0, screen[1]), (0, screen[0])))
        return dense[0]

    def rollup(self, *vals):
        return reduce(lambda x, y: x+y,  vals)


# TODO: IS there a faster option for the coder?  'vectorize' is only
#       a little better than a list comprehension
class PointCountCategories(ar.Aggregator):
    def aggregate(self, glyphset, info, screen):
        points = glyphset.points()
        coder = np.vectorize(info)
        coded = coder(glyphset.data())
        coded = np.array(coded)
        cats = coded.max()

        (width, height) = screen
        dims = (height, width, cats+1)

        # Need cats+1 to do proper categorization...
        ranges = ((0, screen[1]), (0, screen[0]), (0, cats+1))

        data = np.hstack([np.fliplr(points[:, 0:2]), coded[:, np.newaxis]])
        dense = np.histogramdd(data, bins=dims, range=ranges)
        return dense[0]

    def rollup(self, *vals):
        """NOTE: Assumes co-registration of categories..."""
        return reduce(lambda x, y: x+y,  vals)


class Spread(ar.CellShader):
    """
    Spreads the values out in a regular pattern.
    Spreads categories inside their category plane (not between planes).

    TODO: Spread beyond the bounds of the input plot?

    * shape: Shape of the spread.  Supports "rect" and "circle"
    * factor : How wide to spread?
    * anti_alias: Can a value be fractionally spread into a cell
                  (default is false)
    """
    def __init__(self, factor=1, anti_alias=False, shape="circle"):
        self.factor = factor
        self.anti_alias = anti_alias
        self.shape = shape

    def make_k(self):
        """Construct the weights matrix"""
        if self.shape == "rect":
            if not self.anti_alias or (self.factor % 2) == 1:
                kShape = (self.factor+1, self.factor+1)
                k = np.ones(kShape)
            else:
                kShape = (self.factor, self.factor)
                k = np.ones(kShape)
                k[0] = .5
                k[:, 0] = .5
                k[-1] = .5
                k[:, -1] = .5
            return k

        if self.shape == "circle":
            if not self.anti_alias or (self.factor % 2) == 1:
                kShape = (self.factor+1, self.factor+1)
                k = np.ones(kShape)
                r = self.factor//2
                rr = r**2
                for x in range(0, r):
                    for y in range(0, r):
                        if ((x - r)**2 + (y - r)**2) > rr:
                            k[x, y] = 0
                            k[x, -(y+1)] = 0
                            k[-(x+1), -(y+1)] = 0
                            k[-(x+1), y] = 0
                return k
            else:
                kShape = (self.factor+1, self.factor+1)
                k = np.ones(kShape)
                r = self.factor//2
                rr = r**2
                for x in range(0, r):
                    for y in range(0, r):
                        if ((x - r)**2 + (y - r)**2) == rr:
                            k[x, y] = .5
                            k[x, -(y+1)] = .5
                            k[-(x+1), -(y+1)] = .5
                            k[-(x+1), y] = .5
                        elif ((x - r)**2 + (y - r)**2) > rr:
                            k[x, y] = 0
                            k[x, -(y+1)] = 0
                            k[-(x+1), -(y+1)] = 0
                            k[-(x+1), y] = 0
                return k
            return k

    def shade(self, grid):
        k = self.make_k()

        out_dtype = grid.dtype if not self.anti_alias else np.float64
        out = np.empty_like(grid, dtype=out_dtype)
        if len(grid.shape) == 3:
            cats = grid.shape[2]
            for cat in range(cats):
                convolve(grid[:, :, cat], k, output=out[:, :, cat],
                         mode='constant', cval=0.0)
        else:
            convolve(grid, k, mode='constant', cval=0.0, output=out)

        return out


class Log10(ar.CellShader):
    def shade(self, grid):
        mask = (grid == 0)
        out = np.log10(grid)
        out[mask] = 0
        return out


def is_identity_transform(vt):
    return vt == (0, 0, 1, 1)


def load_csv(filename, skip, xc, yc, vc):
    """Turn a csv file into a glyphset.

    This is a fairly naive regulary-expression based parser
    (it doesn't handle quotes, blank lines or much else).
    It is useful for getting simple datasets into the system.
    """
    import re
    source = open(filename, 'r')
    points = []
    data = []

    for i in range(0, skip):
        source.readline()

    for line in source:
        line = re.split("\s*,\s*", line)
        x = float(line[xc].strip())
        y = float(line[yc].strip())
        v = float(line[vc].strip()) if vc >= 0 else 1
        g = [x, y, 0, 0]
        points.append(g)
        data.append(v)

    source.close()
    return Glyphset(np.array(points, order="F"), np.array(data))


def load_hdf(filename, node, xc, yc, vc=None):
    """
    Load a node from an HDF file.

    filename : HDF file to load
    node: Path to relevant HDF table
    xc: Name/index of the x column
    yc: Name/index of the y column
    vc: Name/index of the value column (if applicable)
    cats: List of expected categories.
        If cats is an empty list, a coding will be automatically generated
        Any value not on the list will be assigned category equal to
        list length. This parameter is ignored if vc is not supplied.
    """
    import pandas as pd
    table = pd.read_hdf(filename, node)
    points = table[[xc, yc]]
    a = np.zeros((len(points), 4), order="F")
    a[:, :2] = points

    data = table[vc] if vc else None
    print("Loaded %d items" % len(a))

    return Glyphset(a, data)



from __future__ import print_function
import os

__version__ = "-1.-1.-1"
try:
    version_file = os.path.join(os.path.dirname(__file__), '', '_version.txt')
    with open(version_file, 'r') as version_file:
        __version__ = version_file.read().strip()
except Exception as e:
    print("Error finding version info.")
    print(e)

__version_info__ = dict(zip(["major","minor","micro"], 
                            map(int, __version__.split("."))))



"""
Each info function returns callable that can be used on a single
entry in the dataset.
"""
from __future__ import print_function, division, absolute_import


def const(v):
    """Return the value passed."""
    def f(data):
        return v
    return f


def val(default=None):
    """Return the entire data value. This is the info-version of 'id'.
    """
    def f(data):
        if data is None:
            return default
        else:
            return data
    return f


def valAt(i, default=None):
    """Return the value at a given index in the data part of the input.
         On error returns the default value.
    """
    def f(data):
        try:
            return data[i]
        except:
            return default
    return f


def key(att, default=None):
    "Return the value under a given key in the data part of the input."
    def f(data):
        return data.get(att, default)
    return f


def attribute(att, default=None):
    "Return the value under a given attribute in the data part of the input."
    def f(data):
        return getattr(data, att, default)
    return f


def encode(cats, defcat=-1):
    """Create a function that converts values to numbers.
    The index in the value list passed will be the code-number.
    * cats : Values to create codes for
    * defcat : Default category; defaults to length fo cats list
    """

    if defcat < 0:
        defcat = len(cats)
    codes = dict(zip(cats, xrange(len(cats))))

    def f(val):
        return codes.get(val, defcat)

    return f


class AutoEncode(object):
    """Encoded values as numeric codes.
    Builds up a category->code dictionary incrementally.
    """

    def __init__(self):
        self.next_code = 0
        self.mapping = {}

    def __call__(self, val):
        if val not in self.mapping:
            self.mapping[val] = self.next_code
            self.next_code = self.next_code + 1
        return self.mapping[val]
